
//		CMath.cpp  --  plugin to evaluate math expressions at the command line


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					0
#define VER_BUILD					1
#define VER_PATCH					1

#define PLUGIN_NAME					L"CMath"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif

#define SHORT_BUF_LEN				1024
#define BIG_BUF_LEN					16384


#define CR							0x0d


BOOL NewKeyHook						= FALSE;


const wchar_t PluginFeatures[]		= L"*key";

const wchar_t VarName[]				= L"Ans";
const wchar_t HighlightVarName[]	= L"Highlight";
const wchar_t HighlightKeyName[]	= L"SOFTWARE\\JPPlugins";




BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/cmath.html";
	piInfo.pszDescription	= (LPTSTR) L"Do math at the command line.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"%_4ver";
	unsigned int ver					= 0;


	ExpandVariables( shortbuf, 1 );
	ver = wcstoul( shortbuf, NULL, 10 );
	NewKeyHook = ( ver >= 24 ) && ( ver < 99 );

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%[_transient]%[_pipe]%[noloadmsg]" );
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



int GetRegHighlightVal( HKEY RootKey, LPCTSTR KeyName, LPTSTR outString, DWORD buflen )
{
	if ( !RootKey || !KeyName || !outString || ( buflen < 128 ) )
		return -1;
	else if ( KeyName[0] == '\0' )
		return -1;


	HKEY KeyHandle			= NULL;
	DWORD ValSize			= 0;
	DWORD ValType			= 0;
	int rv					= 0;


	outString[0] = '\0';

	rv = RegOpenKeyEx( RootKey, KeyName, 0, KEY_READ, &KeyHandle );

	if ( rv == ERROR_SUCCESS ) {
		ValSize = buflen * sizeof( wchar_t );
		ValType = 0;

		rv = RegQueryValueEx( KeyHandle, HighlightVarName, NULL, &ValType, (BYTE*) outString, &ValSize );

		if ( rv == ERROR_SUCCESS ) {
			if ( ValType != REG_SZ )
				rv = ERROR_INVALID_DATA;
			else if ( outString[0] == '\0' )
				rv = ERROR_NO_DATA;
		}

		RegCloseKey( KeyHandle );
	}

	if ( rv )
		outString[0] = '\0';

	return rv;
}



BYTE PickHighlightColor( BYTE *Original )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	int fg = 0, bg = 0;
	unsigned int Normal = 0, Inverse = 0;
	int efg = -1, ebg = -1;

	const BYTE FirstColor[16]	= { 0x0f, 0x1f, 0x2f, 0x3f, 0x4f, 0x5f, 0x6f, 0x70, 0x8f, 0x9f, 0xa0, 0xb0, 0xcf, 0xd0, 0xe0, 0xf2 };
	const BYTE SecondColor[16]	= { 0x0e, 0x1b, 0x2b, 0x3e, 0x4b, 0x5b, 0x6b, 0x79, 0x8b, 0x9b, 0xa9, 0xb9, 0xc0, 0xde, 0xec, 0xf0 };


	GetAttribute( &Normal, &Inverse );
	fg = Normal & 0x0f;
	bg = ( Normal & 0xf0 ) >> 4;

	if ( Original )
		*Original = (BYTE) Normal & 0xff;

	if ( fg == bg )
		return 0x07;


	GetEnvironmentVariable( HighlightVarName, temp, SHORT_BUF_LEN );


	if ( temp[0] == '\0' ) {
		wchar_t KeyName[256] = L"";

		swprintf_s( KeyName, 256, L"%s\\%s", HighlightKeyName, PLUGIN_NAME );

		GetRegHighlightVal( HKEY_CURRENT_USER, KeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			GetRegHighlightVal( HKEY_LOCAL_MACHINE, KeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			GetRegHighlightVal( HKEY_CURRENT_USER, HighlightKeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			GetRegHighlightVal( HKEY_LOCAL_MACHINE, HighlightKeyName, temp, SHORT_BUF_LEN );
	}


	if ( temp[0] ) {

		if ( !_wcsnicmp( temp, L"inv", 3 ) )
			ebg = ( Normal & 0x07 ), efg = bg;
		else if ( !_wcsnicmp( temp, L"0x", 2 ) ) {
			unsigned int t = 0;
			wchar_t *sep1 = NULL;

			_set_errno( 0 );
			t = wcstoul( temp + 2, &sep1, 16 );

			if ( !errno && ( *sep1 == '\0' ) )
				efg = t & 0x0f, ebg = ( t & 0xf0 ) >> 4;
		}
		else
			ParseColors( temp, &efg, &ebg );

		if ( ( ebg == 64 ) && ( efg >= 0 ) && ( efg <= 15 ) )		//  kludge:  values 1024 - 1039 treated as 0 - 15,
			ebg = -1;												//    but with default background color.

		if ( ( efg >= 0 ) && ( ebg >= 0 ) )
			if ( ( efg != ebg ) && ( ( efg != fg ) || ( ebg != bg ) ) )
				return ( ebg << 4 ) | efg;

		if ( ( efg >= 0 ) && ( ebg == -1 ) )
			if ( ( efg != fg ) && ( efg != bg ) )
				return ( bg << 4 ) | efg;

		return Normal;
	}

	if ( ( ( bg & 0x08 ) == 0 ) && ( ( fg & 0x08 ) == 0 ) && fg )
		return Normal | 0x08;

	if ( Normal == FirstColor[bg] )
		return SecondColor[bg];
	else
		return FirstColor[bg];
}



int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



DLLExports int WINAPI key( LPKEYINFO ki )
{
	int Key			= ki->nKey;


	if ( !Key && NewKeyHook )
		if ( ki->pszKey )
			if ( !_wcsicmp( ki->pszKey, L"Enter" ) )
				Key = CR;


	if ( ( Key != CR ) || ( ki->pszLine == NULL ) )
		return 0;

	if ( ki->pszLine[0] == '=' )
		ki->pszLine[0] = ':';

	if ( ( ki->pszLine[0] != ':' ) || ( wcslen( ki->pszLine ) >= BIG_BUF_LEN ) )
		return 0;


	wchar_t Exp[BIG_BUF_LEN]			= L"";
	wchar_t Value[BIG_BUF_LEN]			= L"";

	BYTE OldColor	= 0;
	BYTE Highlight	= 0;
	int Row			= 0;
	int Column		= 0;

	
	wcscpy_s( Exp, BIG_BUF_LEN, ki->pszLine + 1 );
	TrimLine( Exp );

	if ( Exp[0] == '\0' )
		return 0;


	CrLf();

	wcscpy_s( Value, BIG_BUF_LEN, Exp );

	if ( wcschr( Value, '%' ) )
		if ( ExpandVariables( Value, 1 ) )
			goto Exit;

	Highlight = PickHighlightColor( &OldColor );


	if ( !Evaluate( Value ) ) {
		Printf( L"%s = ", Exp );
		SetColors( Highlight );
		QPuts( Value );
		SetColors( OldColor );
		CrLf();
		SetEnvironmentVariable( VarName, Value );
	}

Exit:

	//		Update the cursor position so the next prompt doesn't overwrite our output.
	//		Current versions of TCC don't need this, but TCC/LE does.

	GetAbsCursorPosition( &Row, &Column );
	ki->nHomeRow		= Row;
	ki->nHomeColumn		= 0;
	ki->nRow			= Row;
	ki->nColumn			= 0;
	return 0;
}
